﻿using System;
using Microsoft.SPOT;

using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using GTI = Gadgeteer.Interfaces;
using Gralin.NETMF.Nordic;

namespace Gadgeteer.Modules.G2Tech
{
    /// <summary>
    /// A G2Tech.nRF24L01Plus module for Microsoft .NET Gadgeteer
    /// </summary>
    public class G2Tech_nRF24L01Plus : GTM.Module, INRF24L01Plus
    {
        /// <summary>
        /// Module API.
        /// </summary>
        public NRF24L01Plus Api { get; protected set; }

        /// <summary>
        /// Gadgeteer Nordic NRF24L01+ module.
        /// </summary>
        /// <param name="socketNumber">The socket that this module is plugged in to.</param>
        public G2Tech_nRF24L01Plus(int socketNumber)
        {
            var socket = Socket.GetSocket(socketNumber, true, this, null);

            Api = new NRF24L01Plus();
            Api.Initialize(socket.SPIModule, socket.CpuPins[6], socket.CpuPins[4],socket.CpuPins[3]);
            Api.OnDataReceived += onDataReceived;
            Api.OnTransmitFailed += onTransmitFailed;
            Api.OnTransmitSuccess += onTransmitSuccess;
        }

        /// <summary>
        ///   Gets a value indicating whether module is enabled (RX or TX mode).
        /// </summary>
        public bool IsEnabled
        {
            get { return Api.IsEnabled; }
        }

        /// <summary>
        /// Configure the module basic settings. Module needs to be initiaized.
        /// </summary>
        /// <param name="address">RF address (3-5 bytes). The width of this address determins the width of all addresses used for sending/receiving.</param>
        /// <param name="channel">RF channel (0-127)</param>
        public void Configure(byte[] address, byte channel)
        {
            Api.Configure(address, channel);
        }

        /// <summary>
        ///   Enables the module
        /// </summary>
        public void Enable()
        {
            Api.Enable();
        }

        /// <summary>
        ///   Disables the module
        /// </summary>
        public void Disable()
        {
            Api.Disable();
        }

        /// <summary>
        ///   Reads the current rf channel value set in module
        /// </summary>
        /// <returns></returns>
        public byte GetChannel()
        {
            return Api.GetChannel();
        }

        /// <summary>
        ///   Gets the module radio frequency [MHz]
        /// </summary>
        /// <returns>Frequency in MHz</returns>
        public int GetFrequency()
        {
            return Api.GetFrequency();
        }

        /// <summary>
        ///   Send <param name = "bytes">bytes</param> to given <param name = "address">address</param>
        ///   This is a non blocking method.
        /// </summary>
        public void SendTo(byte[] address, byte[] bytes)
        {
            Api.SendTo(address, bytes);
        }

        /// <summary>
        ///   Sends <param name = "bytes">bytes</param> to given <param name = "address">address</param>
        ///   This is a blocking method that returns true if data was received by the recipient or false if timeout occured.
        /// </summary>
        public bool SendTo(byte[] address, byte[] bytes, int timeout)
        {
            return Api.SendTo(address, bytes, timeout);
        }

        /// <summary>
        /// Set one of 6 available module addresses
        /// </summary>
        public void SetAddress(AddressSlot slot, byte[] address)
        {
            Api.SetAddress(slot, address);
        }

        /// <summary>
        /// Read 1 of 6 available module addresses
        /// </summary>
        public byte[] GetAddress(AddressSlot slot, int width)
        {
            return Api.GetAddress(slot, width);
        }
        
        /// <summary>
        ///   Gets module basic status information
        /// </summary>
        public Status GetStatus()
        {
            return Api.GetStatus();
        }

        /// <summary>
        ///   Sets the rf channel value used by all data pipes
        /// </summary>
        /// <param name="channel">7 bit channel value</param>
        public void SetChannel(byte channel)
        {
            Api.SetChannel(channel);
        }

        #region DataReceived event

        /// <summary>
        ///   Occurs when data packet has been received
        /// </summary>
        public event NRF24L01Plus.OnDataRecievedHandler OnDataReceived;
        private NRF24L01Plus.OnDataRecievedHandler _onDataReceived;

        /// <summary>
        /// Raises the <see cref="OnDataReceived"/> event.
        /// </summary>
        /// <param name="data">Received data bytes.</param>  
        protected virtual void onDataReceived(byte[] data)
        {
            if (_onDataReceived == null) _onDataReceived = onDataReceived;
            if (Program.CheckAndInvoke(OnDataReceived, _onDataReceived, data))
                OnDataReceived(data);
        }

        #endregion

        #region TransmitSuccess event

        /// <summary>
        ///   Occurs when ack has been received for send packet
        /// </summary>
        public event NRF24L01Plus.EventHandler OnTransmitSuccess;
        private NRF24L01Plus.EventHandler _onTransmitSuccess;

        /// <summary>
        /// Raises the <see cref="OnTransmitSuccess"/> event.
        /// </summary>
        protected virtual void onTransmitSuccess()
        {
            if (_onTransmitSuccess == null) _onTransmitSuccess = onTransmitSuccess;
            if (Program.CheckAndInvoke(OnTransmitSuccess, _onTransmitSuccess))
                OnTransmitSuccess();
        }

        #endregion

        #region TransmitFailed event

        /// <summary>
        ///   Occurs when no ack has been received for send packet
        /// </summary>
        public event NRF24L01Plus.EventHandler OnTransmitFailed;
        private NRF24L01Plus.EventHandler _onTransmitFailed;

        /// <summary>
        /// Raises the <see cref="OnTransmitFailed"/> event.
        /// </summary>
        protected virtual void onTransmitFailed()
        {
            if (_onTransmitFailed == null) _onTransmitFailed = onTransmitFailed;
            if (Program.CheckAndInvoke(OnTransmitFailed, _onTransmitFailed))
                OnTransmitFailed();
        }

        #endregion
    }
}
